/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.math;

import java.lang.Math;
import java.util.Arrays;

/**
 * A collection of miscellaneous mathematical operations
 * @AUTHOR MorellThomas
 */
public class MiscMath
{
  /**
  * calculates the mean of an array 
  *
  * @param m ~ array
  * @return
  * TODO whats with nans??
  */
  public static double mean(double[] m)
  {
    double sum = 0;
    for (int i = 0; i < m.length; i++)
    {
      if (!Double.isNaN(m[i]))
      {
        sum += m[i];
      }
    }
    return sum / m.length;
  }

  /**
  * calculates the square root of each element in an array
  *
  * @param m ~ array
  *
  * @return
  * TODO
  * make general with function passed -> apply function to each element
  */
  public static double[] sqrt(double[] m)
  {
    double[] sqrts = new double[m.length];
    for (int i = 0; i < m.length; i++)
    {
      sqrts[i] = Math.sqrt(m[i]);
    }
    return sqrts;
  }

  /**
  * calculate element wise multiplication of two arrays
  *
  * @param a ~ array
  * @param b ~ array
  *
  * @return
  */
  public static double[] elementwiseMultiply(byte[] a, double[] b) throws RuntimeException
  {
    if (a.length != b.length)
    {
      throw new SameLengthException(a.length, b.length);
    }
    double[] result = new double[a.length];
    for (int i = 0; i < a.length; i++)
    {
      result[i] = a[i] * b[i];
    }
    return result;
  }
  public static double[] elementwiseMultiply(double[] a, double[] b) throws RuntimeException
  {
    if (a.length != b.length)
    {
      throw new SameLengthException(a.length, b.length);
    }
    double[] result = new double[a.length];
    for (int i = 0; i < a.length; i++)
    {
      result[i] = a[i] * b[i];
    }
    return result;
  }
  public static byte[] elementwiseMultiply(byte[] a, byte[] b) throws RuntimeException
  {
    if (a.length != b.length)
    {
      throw new SameLengthException(a.length, b.length);
    }
    byte[] result = new byte[a.length];
    for (int i = 0; i < a.length; i++)
    {
      result[i] = (byte) (a[i] * b[i]);
    }
    return result;
  }

  /**
  * calculate element wise division of two arrays
  *
  * @param a ~ array
  * @param b ~ array
  *
  * @return
  */
  public static double[] elementwiseDivide(double[] a, double[] b) throws RuntimeException
  {
    if (a.length != b.length)
    {
      throw new SameLengthException(a.length, b.length);
    }
    double[] result = new double[a.length];
    for (int i = 0; i < a.length; i++)
    {
      result[i] = a[i] / b[i];
    }
    return result;
  }

  /**
  * returns true if two arrays are element wise within a tolerance
  *
  * @param a ~ array
  * @param b ~ array
  * @param rtol ~ relative tolerance
  * @param atol ~ absolute tolerance
  * @param equalNAN ~ whether NaN at the same position return true
  *
  * @return
  */
  public static boolean allClose(double[] a, double[] b, double rtol, double atol, boolean equalNAN)
  {
    boolean areEqual = true;
    for (int i = 0; i < a.length; i++)
    {
      if (equalNAN && (Double.isNaN(a[i]) && Double.isNaN(b[i])))
      {
	continue;
      }
      if (Math.abs(a[i] - b[i]) > (atol + rtol * Math.abs(b[i])))
      {
	areEqual = false;
	break;
      }
    }
    return areEqual;
  }

  /**
  * returns the index of the maximum and the maximum value of an array
  * 
  * @pararm a ~ array
  *
  * @returns
  */
  public static int[] findMax(int[] a)
  {
    int max = 0;
    int maxIndex = 0;
    for (int i = 0; i < a.length; i++)
    {
      if (a[i] > max)
      {
	max = a[i];
	maxIndex = i;
      }
    }
    return new int[]{maxIndex, max};
  }
}
